"""
Calibrates the gripper and saves the new motor offsets in motor_config.yaml.
"""

# Standard
import argparse

# Custom
import mimic_hand_middleware


def update_calibration_with_current_pose(
    calibration_path: str = 'git/mimic_hand_middleware/mimic_hand_middleware/motor_config.yaml',
):
    """
    Updates the calibration file at calibration_path with the current position of the
    finger joints as zero position.
    """
    print(f'Got calibration path: {calibration_path}')
    gripper_controller = mimic_hand_middleware.GripperController(
        calibrate_at_start=True, motor_config_yml_path=calibration_path
    )
    gripper_controller.disconnect_motors()


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description="Calibrate the mimic hand's zero position."
    )
    parser.add_argument(
        'calib_yaml_path', type=str, help='Path to the calibration file'
    )
    args = parser.parse_args()
    mimic_hand_middleware.set_motors_to_freerunning_mode()
    input('Move motors to the new zero position, then press enter to save')
    update_calibration_with_current_pose(calibration_path=args.calib_yaml_path)



"""
The following script recalculates multipliers for simple linear + offset joint
kinematics in case the joint limits are violated.
All limits are in degrees.
"""

# Third-party
import numpy as np


def get_new_limits(
    old_angle_min: float,
    old_angle_max: float,
    new_angle_min: float,
    new_angle_max: float,
    old_angle_offset: float,
    old_angle_multiplier: float,
) -> list:
    """
    Returns the new multiplier and offset that match the same endpoints with the new
    angle commands.
    """
    old_angle_max = np.deg2rad(old_angle_max)
    old_angle_min = np.deg2rad(old_angle_min)
    new_angle_max = np.deg2rad(new_angle_max)
    new_angle_min = np.deg2rad(new_angle_min)
    t_min = old_angle_multiplier * (old_angle_offset + old_angle_min)
    t_max = old_angle_multiplier * (old_angle_offset + old_angle_max)
    new_angle_offset = (t_min * new_angle_max - t_max * new_angle_min) / (t_max - t_min)
    new_angle_multiplier = t_max / (new_angle_max + new_angle_offset)
    return new_angle_multiplier, new_angle_offset


def test_limits(
    old_angle_min: float,
    old_angle_max: float,
    new_angle_min: float,
    new_angle_max: float,
    old_angle_offset: float,
    old_angle_multiplier: float,
    new_angle_offset: float,
    new_angle_multiplier: float,
    epsilon: float = 0.0001,
) -> bool:
    """
    Tests if the tendon cmds are the same at the end stops for the old and new
    multiplier and offset settings
    """
    old_angle_max = np.deg2rad(old_angle_max)
    old_angle_min = np.deg2rad(old_angle_min)
    new_angle_max = np.deg2rad(new_angle_max)
    new_angle_min = np.deg2rad(new_angle_min)
    # old tendon endstop
    t_min_old = old_angle_multiplier * (old_angle_offset + old_angle_min)
    t_max_old = old_angle_multiplier * (old_angle_offset + old_angle_max)
    # new tendon endstop
    t_min_new = new_angle_multiplier * (new_angle_offset + new_angle_min)
    t_max_new = new_angle_multiplier * (new_angle_offset + new_angle_max)
    print('Delta min: ', t_min_new - t_min_old)
    print('Delta max: ', t_max_new - t_max_old)
    return (
        np.abs(t_min_old - t_min_new) < epsilon
        and np.abs(t_max_old - t_max_new) < epsilon
    )


if __name__ == '__main__':
    # get constants
    old_angle_min = float(input('old_angle_min'))
    old_angle_max = float(input('old_angle_max'))
    new_angle_min = float(input('new_angle_min'))
    new_angle_max = float(input('new_angle_max'))
    old_angle_offset = float(input('old_angle_offset'))
    old_angle_multiplier = float(input('old_angle_multiplier'))
    new_angle_multiplier, new_angle_offset = get_new_limits(
        old_angle_min,
        old_angle_max,
        new_angle_min,
        new_angle_max,
        old_angle_offset,
        old_angle_multiplier,
    )
    # test
    conversion_success = test_limits(
        old_angle_min,
        old_angle_max,
        new_angle_min,
        new_angle_max,
        old_angle_offset,
        old_angle_multiplier,
        new_angle_offset,
        new_angle_multiplier,
    )
    print(f'mul: {new_angle_multiplier}, offset: {new_angle_offset}')
    print(f'test success: {conversion_success}')



import numpy as np

from .finger_kinematics import FingerKinematics


class SimpleThumbKinematics(FingerKinematics):
    """
    A class that represents the simplified kinematics of the thumb joints.
    """

    def __init__(self, thumb_joint_dict) -> None:
        super().__init__()
        # load simple init constants
        self._abd_multiplier = thumb_joint_dict['simple']['abd_multiplier']
        self._cmc_multiplier = thumb_joint_dict['simple']['cmc_multiplier']
        self._pp_multiplier = thumb_joint_dict['simple']['pp_multiplier']
        self._pp_offset = thumb_joint_dict['simple']['pp_offset']
        self._pip_multiplier = thumb_joint_dict['simple']['pip_multiplier']
        self._pip_offset = thumb_joint_dict['simple']['pip_offset']
        self._cmc_offset = thumb_joint_dict['simple']['cmc_offset']
        self._abd_offset = thumb_joint_dict['simple']['abd_offset']

    def get_tendon_length_from_angle(self, angles: np.ndarray) -> np.ndarray:
        """
        Derive tendon length from the desired angle
        angles is assumed to be of shape (n_joints), in radians.
        Returns the desired tendon length change from the zero position, in
        meters.
        """
        angles = angles.reshape((-1,))
        assert angles.shape[0] == 4, (
            'The shape of the commanded thumb'
            + f'array has to be castable to (4,). It is currently {angles.shape}'
        )
        tendon_cmds = np.zeros_like(angles)
        tendon_cmds[0] = (angles[0] + self._cmc_offset) * self._cmc_multiplier
        tendon_cmds[1] = (angles[1] + self._abd_offset) * self._abd_multiplier
        tendon_cmds[2] = (angles[2] + self._pip_offset) * self._pip_multiplier
        tendon_cmds[3] = (angles[3] + self._pp_offset) * self._pp_multiplier
        return tendon_cmds



__version__ = '0.1.0'
__author__ = 'Ben Forrai'
__credits__ = 'Mimic Robotics AG'
import mimic_hand_api

from mimic_hand_middleware.calibrate_gripper import update_calibration_with_current_pose
from mimic_hand_middleware.conversion_utils import p04
from mimic_hand_middleware.gripper_controller import GripperController
from mimic_hand_middleware.kinematics.mcp_kinematics import MCPKinematics
from mimic_hand_middleware.kinematics.mcp_kinematics_simple import SimpleMCPKinematics
from mimic_hand_middleware.kinematics.pip_kinematics import PIPKinematics
from mimic_hand_middleware.kinematics.thumb_kinematics import SimpleThumbKinematics
from mimic_hand_middleware.set_motors_to_freerunning_mode import (
    set_motors_to_freerunning_mode,
)



"""
Sets the motors to freerunning mode for calibration
"""

from mimic_hand_api import RP2040API as DriverAPI


def set_motors_to_freerunning_mode():
    """
    Moves the hand through the low level API.
    """
    client = DriverAPI()
    client.connect_all_motors()
    client.set_all_motors_to_freerunning_mode()
    # Disconnect motors
    client.disconnect_all_motor_boards()


if __name__ == '__main__':
    set_motors_to_freerunning_mode()



"""
Contains an abstract class for representing finger kinematics and dynamics.
"""

import numpy as np


class FingerKinematics:
    def __init__(self) -> None:
        pass

    def get_tendon_length_from_angle(self, angles: np.ndarray) -> np.ndarray:
        """
        Calculates the tendon length(s) from the given joint angle(s) along a
        batched dimension.
        """
        return angles

    def get_angle_from_tendon_length(self, tendons: np.ndarray) -> np.ndarray:
        """
        Calculates the joint angle(s) from the given tendon length(s) along a
        batched dimension.
        """
        return tendons



# Standard imports
from copy import deepcopy

# Third-party
import numpy as np

# Custom
from .finger_kinematics import FingerKinematics


class SimpleMCPKinematics(FingerKinematics):
    """
    A class for representing a simplified version of MCP kinematics of P04
    The simplified version merely takes the mean and the difference of the
    two opposing tendons and uses them to control the MCP flexion and abduction
    respectively.
    """

    def __init__(self, mcp_joint_init_dict: dict, mcp_base_dimension: int = 4) -> None:
        super().__init__()
        self._base_dim = mcp_base_dimension
        self._abd_multiplier = mcp_joint_init_dict['simple']['abduction_multiplier']
        self._flex_multiplier = mcp_joint_init_dict['simple']['flexion_multiplier']
        # get tendon lengths in the center pos of the motors
        self._tendon_lengths_at_zero = np.zeros(mcp_base_dimension * 2)
        self._tendon_lengths_at_zero = self.get_tendon_length_from_angle(
            np.zeros((mcp_base_dimension, 2))
        )

    def get_tendon_length_from_angle(
        self,
        angles: np.ndarray,
        switch_middle_ring: bool = False,
        swap_dir_on_mcp_abd: bool = False,
    ) -> np.ndarray:
        """
        Derive tendon length from the desired angle
        angles is assumed to be of shate (n_joints, 2), with its first column
        being used for abduction angles (psi) and its second column used for
        flexion angles (omega). Both are in randians.
        """
        assert angles.shape[0] == self._base_dim, (
            f'The shape of the angles {angles.shape} is incompatible with'
            + ' the shape of the transform matrices stack '
            + f' {self._transform_matrices_B_to_T.shape}!'
        )
        # switch ring/middle if specified
        if switch_middle_ring:
            switched_angles = deepcopy(angles)
            angles[-2, :] = switched_angles[-3, :]
            angles[-3, :] = switched_angles[-2, :]
        abd_angles = angles[:, 0]
        if swap_dir_on_mcp_abd:
            # invert all abd angles - motor flipped
            abd_angles *= -1
        flex_angles = angles[:, 1]
        deltas = self._abd_multiplier * abd_angles
        means = self._flex_multiplier * flex_angles
        tendon_lengths = np.zeros_like(self._tendon_lengths_at_zero)
        tendon_lengths[::2] = means - deltas
        tendon_lengths[1::2] = means + deltas
        tendon_lengths -= self._tendon_lengths_at_zero
        return tendon_lengths



"""
The workhorse of hand development: grasp gui allows motor and joint-level
commands through a simple graphical UI for ease of development.
options: --debug, --calibrate, --sim

A modified version of the P0 grasp GUI from ETH SRL, cleaned up and adapted to
a class-based logic.
"""

import copy
import tkinter as tk
from tkinter import NORMAL, messagebox, ttk

import numpy as np

from mimic_hand_middleware import GripperController


class GraspGui(tk.Frame):
    """
    A simple graphical UI for commanding joint/motor angles for the P04
    prototype.
    """

    def __init__(self, parent, *args, **kwargs):
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.gui_root = parent
        self.gc = GripperController(
            '/dev/ttyUSB0', max_motor_current=60, use_sim_motors=False
        )
        self.gc.connect_motors()
        self.gc.enable_torque()
        self.gc.init_joints(calibrate=False)
        self.motor_cmd_array = np.zeros((self.gc.num_of_motors))
        self.joint_cmd_array = np.zeros((self.gc.num_of_joints))
        self.gc.command_motor_angles(self.motor_cmd_array)
        self._setup_gui()

    def _setup_gui(self) -> None:
        """
        Starts the GUI application for the Grasp GUI.
        """
        # setup main application
        self.gui_root.geometry('600x550')
        self.gui_root.resizable(True, True)
        self.gui_root.title('Joint-level control & grasping GUI')
        self.gui_root.protocol('WM_DELETE_WINDOW', self._closing_protocol)
        # set up GUI widget by widget
        self._setup_mode_selection_gui(row_idx=0)
        self._setup_action_buttons(row_idx=2)
        self._setup_motor_controller_gui(row_idx=3)

    def _setup_mode_selection_gui(self, row_idx: int = 0) -> None:
        """
        Sets up the mode selection GUI consisting of a list of RadioButtons.
        The following modes can be selected:
            - Motor Control: for the motor-level commands. The user can use
            sliders to give position commands to the motors, each in a custom
            range.
            - Joint Control (TODO):
            - Grasp Control: allows loading recorded arrays of joint commands
            to the robot (TODO)
        """
        # GUI for settings
        self._control_mode_names = [
            'Motor Control',
            'Joint Control',
            'Grasp Control',
        ]
        self._n_control_modes = len(self._control_mode_names)
        self._mode_selection_col_step = 2
        self._mode_selection_col_span = (
            self._n_control_modes * self._mode_selection_col_step
        )
        ttk.Label(self.gui_root, text='Select Control Mode: ').grid(
            row=row_idx, column=1, columnspan=self._mode_selection_col_span
        )
        self._ctrl_mode = tk.IntVar(master=self.gui_root, value=0)
        self._radio_buttons = []
        row_idx += 1
        for mode_idx, mode_name in enumerate(self._control_mode_names):
            self._radio_buttons.append(
                ttk.Radiobutton(
                    self.gui_root,
                    text=mode_name,
                    variable=self._ctrl_mode,
                    value=mode_idx,
                    command=self._change_control_mode,
                )
            )
            self._radio_buttons[-1].grid(
                row=row_idx,
                column=self._mode_selection_col_step * mode_idx,
                columnspan=self._mode_selection_col_step,
            )

    def _setup_action_buttons(self, row_idx: int = 1) -> None:
        """
        Adds buttons that perform simple callbacks (like resetting motors or
        joints) upon being clicked.
        """
        self._action_button_names = ['Reset Motor Init']
        self._action_button_callbacks = [self._update_motor_init]
        self._action_buttons = []
        self._n_action_buttons = len(self._action_button_names)
        self._action_button_col_step = 2
        self._action_button_span = self._n_action_buttons * self._action_button_col_step
        for button_idx, button in enumerate(
            zip(self._action_button_names, self._action_button_callbacks)
        ):
            button_name = button[0]
            button_callback = button[1]
            action_button = ttk.Button(
                self.gui_root,
                text=button_name,
                command=button_callback,
                state=NORMAL,
            )
            action_button.grid(
                row=row_idx,
                column=self._action_button_col_step * button_idx,
                columnspan=self._action_button_span,
            )
            self._action_buttons.append(action_button)

    def _update_motor_init(self) -> None:
        """
        Recalibrates the motor offset for the GripperController
        """
        self.gc.get_new_motor_zero_positions()
        # set sliders back and reset cmd
        for slider_idx in range(len(self.motor_sliders)):
            self.motor_sliders[slider_idx].set(0)
        self.motor_cmd_array = np.zeros_like(self.motor_cmd_array)

    def _change_control_mode(self) -> None:
        """
        Changes the control mode of the grasp GUI between motor, joint and
        grasp control.
        """
        control_mode_display_name = self._control_mode_names[self._ctrl_mode.get()]
        control_mode_method_name = self._get_control_method_name_from_display_name(
            control_mode_display_name
        )
        # Destroy all other displays
        for display_name in self._control_mode_names:
            if display_name != control_mode_display_name:
                method_name = self._get_control_method_name_from_display_name(
                    display_name
                )
                clearup_fcn = getattr(self, f'_clear_{method_name}ler_gui', None)
                if clearup_fcn is not None:
                    clearup_fcn()
        # Create display
        setup_fcn = getattr(self, f'_setup_{control_mode_method_name}ler_gui', None)
        if setup_fcn is not None:
            setup_fcn()

    def _get_control_method_name_from_display_name(self, display_name: str) -> str:
        """
        Returns the name of the control method that is used for class functions
        from the display name of the control method, eg.
        Motor Control -> motor_control
        """
        method_name = display_name.lower().replace(' ', '_')
        return method_name

    def _setup_motor_controller_gui(self, row_idx=3) -> None:
        """
        Creates sliders for all of the motors.
        """
        # GUI for motor angle control
        controller_row_idx = row_idx
        self.motor_sliders = []
        self.motor_status_displays = []
        self.motor_pos_displays = []
        self.motor_cur_displays = []
        self.motor_labels = []
        slid_id = 0
        main_label = ttk.Label(self.gui_root, text='Motor Control:')
        self.motor_labels.append(main_label)
        main_label.grid(row=7, column=1, columnspan=2)
        for motor_idx in range(self.gc.num_of_motors):
            # make labels
            self.motor_labels.append(
                ttk.Label(self.gui_root, text=f'Motor {motor_idx}')
            )
            self.motor_labels[-1].grid(row=controller_row_idx + motor_idx, column=0)
            # make slider
            range_min = self.gc.motors_limit_lower
            range_max = self.gc.motors_limit_higher
            self.motor_sliders.append(
                ttk.Scale(
                    self.gui_root,
                    from_=range_min,
                    to=range_max,
                    command=lambda value,
                    id=slid_id: self._changed_motor_slider_value_callback(id, value),
                    length=300,
                )
            )
            self.motor_sliders[-1].grid(row=row_idx, column=1, columnspan=2)
            slid_id += 1
            # make motor_cmd displays
            self.motor_status_displays.append(ttk.Label(self.gui_root, text='0.'))
            self.motor_status_displays[-1].grid(row=row_idx, column=3)
            # make motor_pos displays
            self.motor_pos_displays.append(ttk.Label(self.gui_root, text='0.'))
            self.motor_pos_displays[-1].grid(row=row_idx, column=4)
            # make motor current displays
            self.motor_cur_displays.append(ttk.Label(self.gui_root, text='0.'))
            self.motor_cur_displays[-1].grid(row=row_idx, column=5)
            row_idx += 1

    def _setup_joint_controller_gui(self, row_idx=3) -> None:
        """
        Creates sliders for all of the joints.
        """
        # GUI for joint control
        controller_row_idx = row_idx
        self.joint_sliders = []
        self.joint_status_displays = []
        self.joint_pos_displays = []
        self.joint_cur_displays = []
        self.joint_labels = []
        slid_id = 0
        main_label = ttk.Label(self.gui_root, text='Joint Control:')
        self.joint_labels.append(main_label)
        main_label.grid(row=7, column=1, columnspan=2)
        for joint_idx in range(self.gc.num_of_joints):
            # make labels
            joint_name = self.gc.joint_names[joint_idx]
            self.joint_labels.append(ttk.Label(self.gui_root, text=joint_name))
            self.joint_labels[-1].grid(row=controller_row_idx + joint_idx, column=0)
            # make slider
            range_min = self.gc.joint_limit_lower[joint_idx]
            range_max = self.gc.joint_limit_higher[joint_idx]
            self.joint_sliders.append(
                ttk.Scale(
                    self.gui_root,
                    from_=range_min,
                    to=range_max,
                    command=lambda value,
                    id=slid_id: self._changed_joint_slider_value_callback(id, value),
                    length=300,
                )
            )
            self.joint_sliders[-1].grid(row=row_idx, column=1, columnspan=2)
            slid_id += 1
            # make motor_cmd displays
            self.joint_status_displays.append(ttk.Label(self.gui_root, text='0.'))
            self.joint_status_displays[-1].grid(row=row_idx, column=3)
            # make motor_pos displays
            self.joint_pos_displays.append(ttk.Label(self.gui_root, text='0.'))
            self.joint_pos_displays[-1].grid(row=row_idx, column=4)
            # make motor current displays
            self.joint_cur_displays.append(ttk.Label(self.gui_root, text='0.'))
            self.joint_cur_displays[-1].grid(row=row_idx, column=5)
            row_idx += 1

    def _clear_motor_controller_gui(self) -> None:
        """
        Clears all widgets that are related to the motor controller
        """
        motor_control_widgets = [
            'motor_sliders',
            'motor_status_displays',
            'motor_pos_displays',
            'motor_cur_displays',
            'motor_labels',
        ]
        for widget_name in motor_control_widgets:
            if hasattr(self, widget_name):
                widget = getattr(self, widget_name)
                self._destroy_widget_recursively(widget)

    def _destroy_widget_recursively(self, widget) -> None:
        """
        Destroys a multi-level list of widgets recursively
        """
        if type(widget) is list:
            for subwidget in widget:
                self._destroy_widget_recursively(subwidget)
        else:
            widget.destroy()

    def _map_mcp_delta_mean_to_angles(self, cmd: np.ndarray) -> np.ndarray:
        """
        Maps a motor angle command that represents MCP angles as delta and mean
        values to the conventional API format.
        Both the in- and output commands are represented by numpy arrays of
        shape (self._num_of_motors,)
        """
        backup_cmd = copy.deepcopy(cmd)
        cmd[4] = backup_cmd[4] + 0.5 * backup_cmd[5]
        cmd[5] = backup_cmd[4] - 0.5 * backup_cmd[5]
        # return cmd @ self.gc.motors_mcp_to_mean_delta
        return cmd

    def _changed_motor_slider_value_callback(self, id: int, value: float) -> None:
        """
        Updates the corresponding motor value and sends the motor command
        through the gripper controller interface.
        """
        self.motor_cmd_array[id] = value
        self.gc.command_motor_angles(self.motor_cmd_array)
        motor_status = self.gc.get_motor_pos_vel_cur()
        self._update_motor_pos_displays(motor_pos_array=motor_status[0])
        self._update_motor_cur_displays(motor_cur_array=motor_status[2])

    def _changed_joint_slider_value_callback(self, id: int, value: float) -> None:
        """
        Updates the corresponding joint value and sends the motor command
        through the gripper controller interface.
        """
        self.joint_cmd_array[id] = value
        self.gc.command_joint_angles(self.joint_cmd_array)
        motor_status = self.gc.get_motor_pos_vel_cur()
        self._update_motor_cmd_displays(
            status_displays=self.joint_status_displays,
            status_array=self.joint_cmd_array,
        )
        self._update_motor_pos_displays(
            motor_pos_array=motor_status[0], pos_displays=self.joint_pos_displays
        )
        self._update_motor_cur_displays(
            motor_cur_array=motor_status[2], cur_displays=self.joint_cur_displays
        )

    def _update_motor_cmd_displays(
        self, status_displays=None, status_array=None
    ) -> None:
        """
        Updates the commanded motor angle display widgets
        """
        if status_displays is None:
            status_displays = self.motor_status_displays
        if status_array is None:
            status_array = self.motor_cmd_array
        for idx, display_label in enumerate(status_displays):
            display_label.config(
                text=np.format_float_positional(status_array[idx], precision=2)
            )

    def _update_motor_pos_displays(
        self, motor_pos_array=None, pos_displays=None
    ) -> None:
        """
        Updates the motor angle display widgets with motor_pos_array (either a
        numpy array of shape (self.num_of_motors,) or None)
        """
        if motor_pos_array is None:
            motor_pos_array = self.gc.get_motor_pos_vel_cur()[0]
        if pos_displays is None:
            pos_displays = self.motor_pos_displays
        for idx, display_label in enumerate(pos_displays):
            display_label.config(
                text=np.format_float_positional(motor_pos_array[idx], precision=2)
            )

    def _update_motor_cur_displays(
        self, motor_cur_array=None, cur_displays=None
    ) -> None:
        """
        Updates the motor current display widgets
        """
        if motor_cur_array is None:
            motor_cur_array = self.gc.get_motor_pos_vel_cur()[2]
        if cur_displays is None:
            cur_displays = self.motor_cur_displays
        for idx, display_label in enumerate(cur_displays):
            display_label.config(
                text=np.format_float_positional(motor_cur_array[idx], precision=2)
            )

    def _closing_protocol(self):
        """
        Runs upon closing the grasp GUI.
        - Closes the window
        - Disconnects from the motors
        """
        if messagebox.askokcancel('Quit', 'Do you want to quit?'):
            self.gui_root.destroy()
            self.gc.disconnect_motors()


if __name__ == '__main__':
    root = tk.Tk()
    GraspGui(root)
    root.mainloop()



"""
Palm processing script for the Fusion360 SDK python plugin
Reads the parameters describing the finger base transforms wrt. the palm
from the Fusion user-defined parameters. The expected format of the
parameters is the following:
{finger_initial}_{X/Y/Z}_{rot/trans}

Kudos to Ueli Mauer for the original code
Ported from SRL's faive-integration by Ben Forrai (ben.forrai@mimicrobotics.com)
"""

import traceback

import adsk.cam
import adsk.core
import adsk.fusion
import numpy as np
import yaml
from scipy.spatial.transform import Rotation as R


def run(context):
    ui = None
    try:
        app = adsk.core.Application.get()
        ui = app.userInterface
        design = app.activeProduct

        # Set styles of file dialog.
        fileDlg = ui.createFileDialog()
        fileDlg.title = 'Choose location to save YAML file'
        fileDlg.filter = '*.yaml'

        # Show folder dialog
        dlgResult = fileDlg.showSave()
        if dlgResult == adsk.core.DialogResults.DialogOK:
            filepath = fileDlg.filename
        else:
            return

        # Fusion naming convention is defined here
        fingers = ['Thumb', 'Index', 'Middle', 'Ring', 'Pinky']
        cad_finger_naming = {
            'Thumb': 'T',
            'Index': 'I',
            'Middle': 'M',
            'Ring': 'R',
            'Pinky': 'P',
        }
        axes = ['X', 'Y', 'Z']
        moves = ['rot', 'trans']

        # Fusion rotation convention is defined here
        euler_convention = 'ZYX'
        # Init data dict
        data = {}
        # Loop through user parameters and read the euler parameters and translations
        for finger in fingers:
            finger_key = cad_finger_naming[finger]
            for move in moves:
                if move == 'rot':
                    rot_euler = [
                        design.userParameters.itemByName(
                            f'{finger_key}_{axis}_{move}'
                        ).value
                        for axis in euler_convention
                    ]
                    rot_euler = np.asarray(rot_euler)
                    # create scipy rotation object
                    Rot = R.from_euler(
                        euler_convention.lower(), rot_euler, degrees=False
                    )
                elif move == 'trans':
                    translation = [
                        design.userParameters.itemByName(
                            f'{finger}_{axis}_{move}'
                        ).value
                        for axis in axes
                    ]
                    translation = np.asarray(translation)

            data[finger] = {
                'base_rot': Rot.as_quat().tolist(),
                'base_trans': translation.tolist(),
            }

        # Write file to chosen folder
        with open(filepath, 'w') as file:
            yaml.dump(data, file, default_flow_style=False)

        ui.messageBox('Parameters written to YAML file')

    except:
        if ui:
            ui.messageBox('Failed:/n{}'.format(traceback.format_exc()))



"""
Implements the joint-to-tendon length kinematics of the P04 prototype.
"""

import copy
import os

import numpy as np
import yaml

from ..conversion_utils import p04
from .mcp_kinematics_simple import SimpleMCPKinematics

# Custom
from .pip_kinematics import PIPKinematics
from .thumb_kinematics import SimpleThumbKinematics


class HandKinematics:
    """
    Implements the full angle-to-tendon kinematics of the hand.
    """

    def __init__(self, yaml_name: str = 'p_0_4.yaml') -> None:
        """
        Initializes the pip, mcp and thumb kinematics models based on the yaml
        parameter file of the hand.
        """
        self.yaml_path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            '../hand_definitions/yaml',
            yaml_name,
        )
        with open(self.yaml_path) as yaml_file:
            hand_kinematic_params = yaml.load(yaml_file, Loader=yaml.SafeLoader)
        self._pip_kinematics = PIPKinematics(
            x=hand_kinematic_params['pip_joint_values']['x'],
            y=hand_kinematic_params['pip_joint_values']['y'],
            r=hand_kinematic_params['pip_joint_values']['r'],
            scale=hand_kinematic_params['pip_joint_values']['scale'],
        )
        self._mcp_kinematics = SimpleMCPKinematics(
            hand_kinematic_params['mcp_joint_values']
        )
        self._thumb_kinematics = SimpleThumbKinematics(
            hand_kinematic_params['thumb_joint_values']
        )
        self._pip_joint_idxes = p04.PIP_JOINT_IDXES
        self._pip_joint_idxes_swapped = copy.deepcopy(self._pip_joint_idxes)
        self._pip_joint_idxes_swapped[1], self._pip_joint_idxes_swapped[2] = (
            self._pip_joint_idxes_swapped[2],
            self._pip_joint_idxes_swapped[1],
        )
        self._mcp_joint_idxes = p04.MCP_JOINT_IDXES
        self._thumb_joint_idxes = p04.THUMB_JOINT_IDXES
        self._num_tendons = sum(
            [
                len(self._pip_joint_idxes),
                len(self._mcp_joint_idxes),
                len(self._thumb_joint_idxes),
            ]
        )
        self.spool_rad = hand_kinematic_params['spool_radius']

    def get_tendon_lengths_m_from_joint_angles_rad(
        self,
        angles_rad: np.ndarray,
        switch_middle_ring_pip: bool = False,
        switch_middle_ring_mcp: bool = False,
        prepend_0: bool = False,
    ) -> np.ndarray:
        """
        Takes in a joint angle array of shape (n_joints,) (in rad), and returns
        an array of the shape (n_motors,) (in meters).
        If prepend_0 is set, a 0 value is added at the beginning of the array
        to account for the extra unused motors for the wrist.
        """
        tendon_lengths = np.zeros((self._num_tendons))
        if switch_middle_ring_pip:
            tendon_lengths[self._pip_joint_idxes_swapped] = (
                self._pip_kinematics.get_tendon_length_from_angle(
                    angles=angles_rad[self._pip_joint_idxes]
                )
            )
        else:
            tendon_lengths[self._pip_joint_idxes] = (
                self._pip_kinematics.get_tendon_length_from_angle(
                    angles=angles_rad[self._pip_joint_idxes]
                )
            )

        # mcp array needs to be reshaped
        mcp_angles = np.zeros((4, 2))
        mcp_angles[:, 0] = angles_rad[self._mcp_joint_idxes][::2]
        mcp_angles[:, 1] = angles_rad[self._mcp_joint_idxes][1::2]
        tendon_lengths[self._mcp_joint_idxes] = (
            self._mcp_kinematics.get_tendon_length_from_angle(
                angles=mcp_angles,
                switch_middle_ring=switch_middle_ring_mcp,
            )
        )
        tendon_lengths[self._thumb_joint_idxes] = (
            self._thumb_kinematics.get_tendon_length_from_angle(
                angles=angles_rad[self._thumb_joint_idxes]
            )
        )
        if prepend_0:
            tendon_lengths = np.insert(tendon_lengths, 0, 0.0, axis=0)
        return tendon_lengths



"""
High level file providing access to the middleware of the hand prototypes from
P04 onwards.

Made by Benedek Forrai (ben.forrai@mimicrobotics.com)
"""

# standard
import os
import time
from copy import deepcopy
from threading import RLock

# third-party
import numpy as np
import yaml
from mimic_hand_api import RP2040API as DriverAPI

# custom
from .conversion_utils import p04 as hand_utils
from .kinematics.hand_kinematics import HandKinematics


class GripperController:
    """
    Middleware level class for controlling mimic's robotic hands for prototypes
    newer than P04.
    Provides the following methods for easier integration:
    - connect_to_motors: connects to the motors of the hand to the API
    - init_joints: "homes" the hands and starts the EKF tracking if specified.
    - command_joint_angles: sends the desired joint (!) angles to the hand. The
        hand then takes a motor configuration that is as close to the commanded
        joint angles as possible.
    - get_joint_pos: returns the current position of the joints (in deg.)
    - get_joint_vel: returns the speed of the joints (in deg./s)
    - get_motor_pos_vel_cur: returns the position (deg), rot. speed (deg/s) and
        current (mA) of the motors
    """

    def __init__(
        self,
        prototype_name: str = 'p4',
        port: str = '/dev/mimic_hand_driver',
        config_yml: str = 'p_0_4.yaml',
        motor_config_yml_path: str = 'git/mimic_hand_middleware/mimic_hand_middleware/motor_config.yaml',
        init_motor_pos_update_thread: bool = True,
        use_sim_joint_measurement: bool = False,
        compliant_test_mode: bool = False,
        max_motor_current: float = 200.0,
        use_sim_motors: bool = False,
        calibrate_at_start: bool = False,
        motor_port_env_var: str = 'HAND_PORT',
    ) -> None:
        """
        The controller has the following parameters:
        - prototype_name (str): name of the hand prototype that is currently in
            use
        - port (str): name of the USB port to connect to
        - config_yml (str): name of config file defining the MuscleGroups in a
            hand. Legacy, kept only for backwards compatibility
        - init_motor_pos_update_thread (bool): legacy, kept only for
            compatibility
        - use_sim_joint_measurement (bool): whether to run the EKF tracker
        - compliant_test_mode (bool): legacy, kept only for compatibility
        (not back-driveable currently sadly)
        - max_motor_current (float): maximum allowed motor current in mA
        - use_sim_motors (bool): if set to true, run a virtual simple sim of
            the hand
        - calibrate_at_start (bool): if set to true, the initial position of the hand
        will be taken as the new zero position for the hand's motors.
        """
        # Init class variables
        self._prototype = prototype_name
        self._cfg_file_name = config_yml
        self._motor_cfg_file = os.path.join(
            os.path.expanduser('~'), motor_config_yml_path
        )
        self._init_pos_update = init_motor_pos_update_thread
        self._sim_measurement = use_sim_joint_measurement
        self._sim_motors = use_sim_motors
        self._compliant_mode_legacy = compliant_test_mode
        self._max_motor_current = max_motor_current
        # Init port for hand communication
        if motor_port_env_var in os.environ.keys():
            self._port = '/dev/' + str(os.environ[motor_port_env_var])
        else:
            self._port = port
        # Import/init low-level utilities
        self._hand_kinematics = HandKinematics(yaml_name=self._cfg_file_name)
        hand_utils.init_matrices()
        # Init constants
        self.num_of_joints = len(hand_utils.JOINT_NAMES_NORMAL)
        self.joint_names = hand_utils.JOINT_NAMES_NORMAL
        self.joint_limit_lower = hand_utils.GC_LIMITS_LOWER
        self.joint_limit_higher = hand_utils.GC_LIMITS_UPPER
        self.num_of_motors = len(hand_utils.MOTOR_NAMES)
        self._motor_zero_offsets = np.zeros((self.num_of_motors,))
        self.motors_limit_lower = hand_utils.MOTOR_LIMIT_LOWER
        self.motors_limit_higher = hand_utils.MOTOR_LIMIT_HIGHER
        self.motors_mcp_to_mean_delta = hand_utils.MCP_TO_MEAN_DELTA
        self.motors_mcp_from_mean_delta = hand_utils.MCP_FROM_MEAN_DELTA
        self.torque_enabled = False
        # Disable scientific notation for printing np arrays
        np.set_printoptions(suppress=True)
        if calibrate_at_start:
            self.connect_motors()
            self._driver_api.set_all_motors_to_freerunning_mode()
            self._calibrate_motor_angles()
        else:
            try:
                print(f'Loading motor cfg: {self._motor_cfg_file}')
                with open(self._motor_cfg_file) as motor_cfg_file:
                    self._motor_config = yaml.safe_load(motor_cfg_file)
                self._motor_zero_offsets = np.array(
                    self._motor_config['zero_offsets']
                ).reshape((-1,))
                print('motor offset: ', self._motor_zero_offsets)
            except FileNotFoundError:
                print('Failed to find motor_config.yaml!')

    def connect_motors(self) -> None:
        """
        Connects to the motors of the hand.
        If self._sim_motors is True, does not do
        anything.
        """
        if not self._sim_motors and not hasattr(self, '_driver_client'):
            self._driver_api = DriverAPI(port_name=self._port)
            self._driver_api.connect_all_motors()
            self._driver_api.set_current_limit(self._max_motor_current)
        else:
            pass

    def disconnect_motors(self) -> None:
        """
        Disconnects motors from all boards.
        If self.sim_motors is True, does not do anything.
        """
        if not self._sim_motors:
            self._driver_api.disconnect_all_motor_boards()
            print('Disconnected all motors. Controller can be shut down now!')
        else:
            pass

    def enable_torque(self) -> None:
        """
        Sets the motors to current-limited positional control mode.
        """
        self._driver_api.set_all_motors_to_cur_lim_pos_control_mode()
        self.torque_enabled = True

    def disable_torque(self) -> None:
        """
        Sets the motors to freerunning mode.
        """
        self._driver_api.set_all_motors_to_freerunning_mode()
        self.torque_enabled = False

    def init_joints(self, calibrate: bool = False) -> None:
        """
        Initializes the internal buffers of the hand and "homes" it to its zero
        position.
        If calibrate is set to True, calibrates the hand by driving it to its
        limit positions and registering motor positions.
        """
        self._motor_lock = RLock()
        self._joint_value_lock = RLock()
        # set zero initial position
        with self._joint_value_lock:
            self._joint_array = np.zeros((self.num_of_joints,))
        if not self._sim_motors:
            if calibrate:
                self._calibrate_motor_angles()
                # enable torque again
                self.enable_torque()

    def _get_motor_angles_from_tendon_lengths(
        self, commanded_lengths: np.ndarray
    ) -> None:
        """
        Get motor angles (uncalibrated) from free tendon lenghts.
        @param commanded_lengths (np.ndarray of shape (num_of_motors,)):
            the commanded lengths (in meters) for each of the 16 tendons.
        @return motor_angles (np.ndarray of shape (num_of_motors,)): motor
            angles that result in the desired tendon lengths, in radians.
        """
        return commanded_lengths / self._hand_kinematics.spool_rad

    def command_joint_angles(
        self, commanded_angles: np.ndarray, convert_from_degrees: bool = True
    ) -> None:
        """
        Commands joint angle command_angles (in degrees) to the low level
        controller. If self._sim_motors is True, this only updates the current
        commanded joint array.
        """
        commanded_angles = deepcopy(commanded_angles)
        commanded_angles = self._filter_joint_angle_limits(commanded_angles)
        if convert_from_degrees:
            commanded_angles *= np.pi / 180
        commanded_angles = commanded_angles.reshape((-1,))
        assert commanded_angles.shape[0] == self.num_of_joints, (
            'Incorrect command dimension; got commanded shape of'
            + f' {commanded_angles.shape} while number of joints is '
            + f'{self.num_of_joints}!'
        )
        if not self._sim_motors:
            free_tendon_lengths = self._get_tendon_lengths_from_angles(commanded_angles)
            raw_spool_angles = self._get_motor_angles_from_tendon_lengths(
                free_tendon_lengths
            )
            self.command_motor_angles(
                commanded_angles=raw_spool_angles, convert_from_radians=True
            )
        else:
            with self._joint_value_lock:
                self._joint_array = commanded_angles

    def _get_tendon_lengths_from_angles(self, angles_rad: np.ndarray) -> np.ndarray:
        """
        Calculates the desired "free" (difference from normal pos.) tendon
        length (in meters) from the desire joint array.
        """
        tendon_lengths = (
            self._hand_kinematics.get_tendon_lengths_m_from_joint_angles_rad(angles_rad)
        )
        return tendon_lengths

    def _get_calibrated_motor_angles(self, raw_spool_angles: np.ndarray) -> np.ndarray:
        """
        Takes uncalibrated motor angle commands (in radians) and shifts them
        with the zero positions measured during calibration.
        """
        motor_angles = raw_spool_angles + self._motor_zero_offsets
        return motor_angles

    def _filter_joint_angle_limits(self, raw_joint_angles: np.ndarray) -> np.ndarray:
        """
        Filters raw_joint_angles (assumed to be in degrees) to stay inside the
        joint limits.
        """
        filtered_joint_angles = np.clip(
            a=raw_joint_angles,
            a_min=self.joint_limit_lower,
            a_max=self.joint_limit_higher,
        )
        return filtered_joint_angles

    def command_motor_angles(
        self, commanded_angles: np.ndarray, convert_from_radians: bool = False
    ) -> None:
        """
        Commands desired motor angles (in degrees)
        """
        cmd_angles = deepcopy(commanded_angles)
        if convert_from_radians:
            cmd_angles = np.rad2deg(cmd_angles)
        cmd_angles = self._get_calibrated_motor_angles(cmd_angles)
        self._driver_api.command_middleware_motor_position_array(cmd_angles)

    def get_joint_pos(self, use_ekf: bool = True) -> np.ndarray:
        """
        Returns the joint position array of the hand in degrees
        """
        joint_array = np.zeros((self.num_of_joints))
        with self._joint_value_lock:
            joint_array = deepcopy(self._joint_array)
        return joint_array

    def get_joint_vel(self) -> np.ndarray:
        """
        Returns the joint velocity array of the hand in degrees/s
        """
        if not self._sim_motors:
            # TODO later this needs to come from the jacobian
            joint_vel_array = np.zeros((self.num_of_joints, 1))
        else:
            joint_vel_array = np.zeros((self.num_of_joints, 1))
        return joint_vel_array

    def get_motor_pos_vel_cur(self) -> np.ndarray:
        """
        Returns the position (degrees), velocity (degrees/s) and current (mA)
        measured by the motors
        """
        motor_pos = np.zeros((self.num_of_motors,))
        motor_vel = np.zeros((self.num_of_motors,))
        motor_cur = np.zeros((self.num_of_motors,))
        if not self._sim_motors:
            motor_cur = self._driver_api.get_motor_currents()
            motor_pos = self._driver_api.get_motor_positions()
        return motor_pos, motor_vel, motor_cur

    def _calibrate_motor_group(
        self,
        motor_group_idxes: np.ndarray,
        forward_max_current_mA: float,
        backward_max_current_mA: float,
        starting_resolution_deg: float = 5.0,
        resolution_refine_steps: int = 2,
        resolution_refine_scale: float = 5.0,
    ) -> list:
        """
        Moves the motors selected by motor_group_idxes (with the middleware
        convention, see README.md) to the joint limits in both directions,
        until the respective current limit (forward/backward_max_current_mA) is
        hit. Repeats the test of the limit with smaller steps than the starting
        step size (starting_resolution_deg), with each step cycle scaled by
        1/resolution_refine_scale.
        """
        print(f'Calibrating motor group idxes: {motor_group_idxes}')
        step_sizes = [
            starting_resolution_deg / (resolution_refine_scale**i)
            for i in range(resolution_refine_steps)
        ]
        # calibrate positive direction
        start_angles = np.zeros_like(motor_group_idxes)
        for step_size in step_sizes:
            self._drive_group_until_limit(
                motor_group_idxes,
                max_current_mA=forward_max_current_mA,
                step=step_size,
                start_angles=start_angles,
            )
            limit_pos = self._driver_api.get_motor_positions()
            start_angles = limit_pos[motor_group_idxes] * 0.75
            time.sleep(0.5)
        forward_limit_pos = limit_pos[motor_group_idxes]
        # calibrate negative direction
        start_angles = np.zeros_like(motor_group_idxes)
        self._driver_api.command_middleware_motor_position_array(
            middleware_cmd_array=np.zeros((self.num_of_motors,))
        )
        # delay for currents to got back to normal
        time.sleep(0.5)
        for step_size in step_sizes:
            self._drive_group_until_limit(
                motor_group_idxes,
                max_current_mA=backward_max_current_mA,
                step=-step_size,
                start_angles=start_angles,
            )
            limit_pos = self._driver_api.get_motor_positions()
            start_angles = deepcopy(limit_pos[motor_group_idxes])
            start_angles *= 0.75
            time.sleep(0.5)
        backward_limit_pos = limit_pos[motor_group_idxes]
        return [forward_limit_pos, backward_limit_pos]

    def _drive_group_until_limit(
        self,
        motor_group_idxes: np.ndarray,
        max_current_mA: float,
        step: float,
        start_angles: np.ndarray = None,
    ) -> None:
        """
        Moves the motors selected by motor_group_idxes (with the middleware
        convention, see README.md) with step until max_current_mA is hit. Sends
        the rest of the motors to 0 deg.
        """
        # init start angle
        motor_group_idxes = motor_group_idxes.reshape(-1)
        cmd_array = np.zeros((self.num_of_motors,))
        if start_angles is not None:
            cmd_array[motor_group_idxes] = start_angles
        # drive motor group to limit from start angle
        driven_group_cmd_array = cmd_array[motor_group_idxes]
        motor_cur = self._driver_api.get_motor_currents()
        while np.any(motor_cur[motor_group_idxes] < max_current_mA):
            driven_group_cmd_array[motor_cur[motor_group_idxes] < max_current_mA] += (
                step
            )
            cmd_array[motor_group_idxes] = driven_group_cmd_array
            self._driver_api.command_middleware_motor_position_array(cmd_array)
            time.sleep(abs(0.02 * step))
            motor_cur = self._driver_api.get_motor_currents()
            print('motor cur: ', motor_cur)

    def _calibrate_motor_angles(self) -> None:
        """
        Reads the zero positions of the motors and writes them to a yaml file.
        """
        self.get_new_motor_zero_positions()
        motor_dict = self._get_motor_dict()
        with open(self._motor_cfg_file, 'w') as cfg_file:
            yaml.safe_dump(
                data=motor_dict,
                stream=cfg_file,
            )
            print('Saved calibration results and applied calibration!')

    def _get_motor_dict(self) -> dict:
        """
        Assembles a motor dict from the current state of the motors that
        we can write to a calibration file.
        """
        motor_zero_offsets = self._motor_zero_offsets
        motor_dict = {
            'zero_offsets': [float(offset) for offset in list(motor_zero_offsets)]
        }
        return motor_dict

    def get_new_motor_zero_positions(self) -> None:
        """
        Recalibrates the offsets for the motors such that the current position
        will be treated as the new zero position during operation.
        """
        # turn motors to freerunning mode
        self._motor_zero_offsets = self._driver_api.get_motor_positions()



"""
Scripts capturing the tendon-joint angle dynamics of the PIP joint on the P04
hand prototype.
"""

import numpy as np

from .finger_kinematics import FingerKinematics


class PIPKinematics(FingerKinematics):
    """
    Class for representing PIP joint kinematics and dynamics
    """

    def __init__(self, x, y, r, scale=1.0) -> None:
        super().__init__()
        self.x = x  # tendon start points in x axis
        self.y = y  # tendon start points along y axis
        self.r = r  # joint rad
        self.scale = scale
        self.tendon_length_at_zero = 0.0
        self.phi_dash = self._get_phi_dash()
        self.max_tangent_len = self._get_maximal_free_tendon_length()
        self.tendon_length_at_zero = self.get_tendon_length_from_angle(np.array([0.0]))

    def _get_phi_dash(self) -> float:
        """
        Calculate the angle phi_dash, which is the point where the tendon no
        longer is tangent to the joint.
        """
        hypotenuse = np.sqrt((self.y + self.r) ** 2 + self.x**2)
        phi_dash_ = np.arccos(self.r / hypotenuse) + np.arcsin(self.x / hypotenuse)
        phi_dash = np.pi - phi_dash_
        return phi_dash

    def _get_maximal_free_tendon_length(self) -> float:
        """
        Calculate the maximal tangent lenght
        """
        max_free = np.sqrt(
            np.square(self.x - self.r * np.sin(self.phi_dash))
            + np.square(self.y + self.r * (1 + np.cos(self.phi_dash)))
        )
        return max_free

    def _get_tendon_length_from_tangent_angle(self, angles: np.ndarray) -> np.ndarray:
        """
        Get tendon lengths from an angle that is smaller than self.phi_dash
        """
        # get the part of the tendon that is tangent to the joint
        angles = self.phi_dash - angles
        # the "free" tendon length is self.max_tangent_len
        tendon_lengths = angles * self.r + self.max_tangent_len
        return tendon_lengths

    def _get_tendon_length_from_large_angle(self, angles: np.ndarray) -> np.ndarray:
        """
        Get tendon lengths from large angles (>self.phi_dash)
        """
        tendon_length = np.sqrt(
            np.square(self.x - self.r * np.sin(angles))
            + np.square(self.y + self.r * (1 + np.cos(angles)))
        )
        return tendon_length

    def get_tendon_length_from_angle(self, angles: np.ndarray) -> np.ndarray:
        """
        Derive tendon length from the desired angle
        """
        tendon_lengths = np.zeros_like(angles)
        # angles smaller than phi_dash
        tendon_lengths[angles < self.phi_dash] = (
            self._get_tendon_length_from_tangent_angle(angles[angles < self.phi_dash])
        )
        # angles larger or equal to phi_dash
        tendon_lengths[angles >= self.phi_dash] = (
            self._get_tendon_length_from_large_angle(angles[angles >= self.phi_dash])
        )
        # TODO remove ugly scaling later
        tendon_lengths *= self.scale
        # shift so we get a length of 0 for a angle of 0
        tendon_lengths = tendon_lengths - self.tendon_length_at_zero
        return tendon_lengths



import numpy as np

from .finger_kinematics import FingerKinematics


class MCPKinematics(FingerKinematics):
    """
    A class for representing MCP kinematics of P04
    """

    def __init__(self, mcp_joint_init_dict: dict, mcp_base_dimension: int = 4) -> None:
        super().__init__()
        self._h_1 = mcp_joint_init_dict['h_1']
        self._h_2 = mcp_joint_init_dict['h_2']
        tendon_position_names = [
            'sfr',
            'sfl',
            'sbr',
            'sbl',
            'mbr',
            'mbl',
            'efr',
            'efl',
            'ebr',
            'ebl',
        ]
        # initialize tendon end and beginning local coords
        for tendon_pos_name in tendon_position_names:
            current_vector = np.array(
                [
                    mcp_joint_init_dict[f'r_{tendon_pos_name}']['x'],
                    mcp_joint_init_dict[f'r_{tendon_pos_name}']['y'],
                    mcp_joint_init_dict[f'r_{tendon_pos_name}']['z'],
                    1.0,
                ]
            ).reshape((1, -1, 1))
            current_vector = np.tile(current_vector, (mcp_base_dimension, 1, 1))
            setattr(self, f'_r_{tendon_pos_name}', current_vector)
        # initialize tendon legnths
        self._front_right_tendon_lengths = np.zeros((mcp_base_dimension, 4))
        self._front_left_tendon_lengths = np.zeros((mcp_base_dimension, 4))
        self._back_right_tendon_lengths_s1 = np.zeros((mcp_base_dimension, 4))
        self._back_left_tendon_lengths_s1 = np.zeros((mcp_base_dimension, 4))
        self._back_right_tendon_lengths_s2 = np.zeros((mcp_base_dimension, 4))
        self._back_left_tendon_lengths_s2 = np.zeros((mcp_base_dimension, 4))
        # initialize transform matrices
        self._transform_matrices_B_to_M = np.zeros((mcp_base_dimension, 4, 4))
        self._transform_matrices_M_to_T = np.zeros((mcp_base_dimension, 4, 4))
        self._transform_matrices_B_to_T = np.zeros((mcp_base_dimension, 4, 4))
        # get tendon lengths in the center pos of the motors
        self._tendon_lengths_at_zero = np.zeros(mcp_base_dimension * 2)
        self._tendon_lengths_at_zero = self.get_tendon_length_from_angle(
            np.zeros((mcp_base_dimension, 2))
        )

    def get_tendon_length_from_angle(self, angles: np.ndarray) -> np.ndarray:
        """
        Derive tendon length from the desired angle
        angles is assumed to be of shate (n_joints, 2), with its first column
        being used for abduction angles (psi) and its second column used for
        flexion angles (omega). Both are in randians.
        """
        assert angles.shape[0] == self._transform_matrices_B_to_T.shape[0], (
            f'The shape of the angles {angles.shape} is incompatible with'
            + ' the shape of the transform matrices stack '
            + f' {self._transform_matrices_B_to_T.shape}!'
        )
        psis = angles[:, 0]
        omegas = angles[:, 1]
        # Initialize transform matrices
        self._update_base_to_middle_tf_from_angle(psis=psis)
        self._update_middle_to_top_tf_from_angle(omegas=omegas)
        self._update_base_to_top_tf_from_subtransforms()
        # Calculate front tendons (modeled as a single line)
        self._front_right_tendon_lengths = (
            self._r_sfr - self._transform_matrices_B_to_T @ self._r_efr
        )
        self._front_left_tendon_lengths = (
            self._r_sfl - self._transform_matrices_B_to_T @ self._r_efl
        )
        # Calculate the back tendons (modeled in two segments)
        # right
        self._back_right_tendon_lengths_s1 = (
            self._r_sbr - self._transform_matrices_B_to_M @ self._r_mbr
        )
        self._back_right_tendon_lengths_s2 = (
            self._transform_matrices_B_to_M @ self._r_mbr
            - self._transform_matrices_M_to_T @ self._r_ebr
        )
        # left
        self._back_left_tendon_lengths_s1 = (
            self._r_sbl - self._transform_matrices_B_to_M @ self._r_mbl
        )
        self._back_left_tendon_lengths_s2 = (
            self._transform_matrices_B_to_M @ self._r_mbl
            - self._transform_matrices_M_to_T @ self._r_ebl
        )
        self._back_left_tendon_lengths_s2 = self._back_right_tendon_lengths_s2.reshape(
            -1, 4, 1
        )
        # calculate lengths and return
        front_right_tendon_lenghts = np.linalg.norm(
            self._front_right_tendon_lengths, axis=1
        )
        front_left_tendon_lenghts = np.linalg.norm(
            self._front_left_tendon_lengths, axis=1
        )
        back_right_tendon_lengths = np.linalg.norm(
            self._back_right_tendon_lengths_s1, axis=1
        ) + np.linalg.norm(self._back_right_tendon_lengths_s2, axis=1)
        back_left_tendon_lengths = np.linalg.norm(
            self._back_left_tendon_lengths_s1, axis=1
        ) + np.linalg.norm(self._back_left_tendon_lengths_s2, axis=1)
        tendon_lengths = self._get_differential_tendon_lengths(
            front_right_tendon_lenghts,
            back_right_tendon_lengths,
            front_left_tendon_lenghts,
            back_left_tendon_lengths,
        )
        tendon_lengths -= self._tendon_lengths_at_zero
        return tendon_lengths

    def _get_differential_tendon_lengths(
        self,
        front_right_tendon_lengths: np.ndarray,
        back_right_tendon_lengths: np.ndarray,
        front_left_tendon_lengths: np.ndarray,
        back_left_tendon_lengths: np.ndarray,
        merge_mode: str = 'alternating',
    ) -> np.ndarray:
        """
        Calculate the sum of the opposing tendons to get the commanded tendon
        legths.
        """
        right_tendons = front_right_tendon_lengths + back_left_tendon_lengths
        left_tendons = front_left_tendon_lengths + back_right_tendon_lengths
        if merge_mode == 'alternating':
            tendon_lengths = np.zeros((2 * right_tendons.shape[0]))
            tendon_lengths[::2] = right_tendons.reshape((-1,))
            tendon_lengths[1::2] = left_tendons.reshape((-1,))
        else:
            raise NotImplementedError
        return tendon_lengths

    def _update_base_to_middle_tf_from_angle(self, psis: np.ndarray):
        """
        Calculates the stacked transform matrices that take the tendon from
        the base to the middle
        """
        self._transform_matrices_B_to_M[:, 0, 0] = 1.0
        self._transform_matrices_B_to_M[:, 0, 1:] = 0.0
        self._transform_matrices_B_to_M[:, 1, 1] = np.cos(psis)
        self._transform_matrices_B_to_M[:, 1, 2] = -np.sin(psis)
        self._transform_matrices_B_to_M[:, 1, 3] = 0.0
        self._transform_matrices_B_to_M[:, 2, 1] = np.sin(psis)
        self._transform_matrices_B_to_M[:, 2, 2] = np.cos(psis)
        self._transform_matrices_B_to_M[:, 2, 3] = self._h_1
        self._transform_matrices_B_to_M[:, 3, :] = 0.0
        self._transform_matrices_B_to_M[:, 3, 3] = 1.0

    def _update_middle_to_top_tf_from_angle(self, omegas: np.ndarray):
        """
        Calculates the stacked transform matrices that take the tendon from
        the middle to the top of the MCP joint
        """
        self._transform_matrices_M_to_T[:, 0, 0] = np.cos(omegas)
        self._transform_matrices_M_to_T[:, 0, 1] = 0.0
        self._transform_matrices_M_to_T[:, 0, 2] = np.sin(omegas)
        self._transform_matrices_M_to_T[:, 0, 3] = 0.0
        self._transform_matrices_M_to_T[:, 1, :] = 0.0
        self._transform_matrices_M_to_T[:, 1, 1] = 1.0
        self._transform_matrices_M_to_T[:, 2, 0] = -np.sin(omegas)
        self._transform_matrices_M_to_T[:, 2, 1] = 0.0
        self._transform_matrices_M_to_T[:, 2, 2] = np.cos(omegas)
        self._transform_matrices_M_to_T[:, 2, 3] = self._h_2
        self._transform_matrices_M_to_T[:, 3, :] = 0.0
        self._transform_matrices_M_to_T[:, 3, 3] = 1.0

    def _update_base_to_top_tf_from_subtransforms(self):
        """
        Calculates the stacked end-to-end transform by combining the two sub-
        transforms (base-to-middle and middle-to-base)
        """
        self._transform_matrices_B_to_T = (
            self._transform_matrices_B_to_M @ self._transform_matrices_M_to_T
        )

    def _get_base_to_top_tf_from_angles(self, psis: np.ndarray, omegas: np.ndarray):
        """
        Calculates the stacked end-to-end transform using the given angles
        Only for checking/debug
        """
        _transform_matrices_B_to_T = np.zeros(
            (self._transform_matrices_B_to_M.shape[0], 4, 4)
        )
        _transform_matrices_B_to_T[:, 0, 0] = np.cos(omegas)
        _transform_matrices_B_to_T[:, 0, 1] = 0.0
        _transform_matrices_B_to_T[:, 0, 2] = np.sin(omegas)
        _transform_matrices_B_to_T[:, 0, 3] = 0.0
        _transform_matrices_B_to_T[:, 1, 0] = np.sin(psis) * np.sin(omegas)
        _transform_matrices_B_to_T[:, 1, 1] = np.cos(psis)
        _transform_matrices_B_to_T[:, 1, 2] = -np.sin(psis) * np.cos(omegas)
        _transform_matrices_B_to_T[:, 1, 3] = -np.sin(psis) * self._h_1
        _transform_matrices_B_to_T[:, 2, 0] = -np.cos(psis) * np.sin(omegas)
        _transform_matrices_B_to_T[:, 2, 1] = np.sin(psis)
        _transform_matrices_B_to_T[:, 2, 2] = np.cos(psis) * np.cos(omegas)
        _transform_matrices_B_to_T[:, 2, 3] = np.cos(psis) * self._h_2 + self._h_1
        _transform_matrices_B_to_T[:, 3, :] = 0.0
        _transform_matrices_B_to_T[:, 3, 3] = 1.0
        return _transform_matrices_B_to_T



"""
Joint-to-motor conversion utilities for the P04 prototype (first one out of
Mimic).
"""

# Standard
import os
from typing import Dict

import defusedxml.ElementTree as ET

# Third-party
import numpy as np

PROTOTYPE_SERIES = 'p4'


def get_urdf_path(prototype: str = 'p4') -> str:
    """
    Gets the path to the kinematic description to the robot hand
    """
    # check is the MIMIC_USE_ROS env variable exists and is set
    if 'MIMIC_USE_ROS' in os.environ:
        from ament_index_python.packages import get_package_share_directory

        resources_path = get_package_share_directory('mimic_viz')
    else:
        script_path = os.path.dirname(os.path.realpath(__file__))
        resources_path = os.path.join(script_path, '../hand_definitions')
    urdf_path = os.path.join(resources_path, f'urdf/{prototype}/converted.urdf')
    return urdf_path


def parse_urdf(urdf_path: str) -> list:
    """
    Get a list of all the joints from a .urdf file
    """
    # get full list from the .urdf file
    joint_names = []
    tree = ET.parse(urdf_path)
    root = tree.getroot()
    for joint in root.iter('joint'):
        joint_name = joint.attrib['name']
        if not joint_name.endswith('offset'):
            joint_names.append(joint_name)
    return joint_names


URDF_PATH = get_urdf_path(PROTOTYPE_SERIES)
JOINT_NAMES_MODEL_P04 = parse_urdf(urdf_path=URDF_PATH)
assert len(JOINT_NAMES_MODEL_P04) == 20  # added dp joints on each finger
JOINT_NAMES_NORMAL_P04 = []
for joint_name in JOINT_NAMES_MODEL_P04:
    if not joint_name.endswith('dp') and not joint_name.endswith('offset'):
        JOINT_NAMES_NORMAL_P04.append(joint_name)
assert len(JOINT_NAMES_NORMAL_P04) == 16

MP2DP_FACTOR = 1.0
GC_TENDONS = {
    'thumb_base2cmc': {},
    'thumb_cmc2mcp': {},
    'thumb_mcp2pp': {},
    'thumb_pp2dpA': {},
    'index_base2mcp': {},
    'index_mcp2pp': {},
    'index_pp2mp': {},
    'index_mp2dp': {'index_mcp2pp': MP2DP_FACTOR},
    'middle_base2mcp': {},
    'middle_mcp2pp': {},
    'middle_pp2mp': {},
    'middle_mp2dp': {'middle_mcp2pp': MP2DP_FACTOR},
    'ring_base2mcp': {},
    'ring_mcp2pp': {},
    'ring_pp2mp': {},
    'ring_mp2dp': {'ring_mcp2pp': MP2DP_FACTOR},
    'pinky_base2mcp': {},
    'pinky_mcp2pp': {},
    'pinky_pp2mp': {},
    'pinky_mp2dp': {'pinky_mcp2pp': MP2DP_FACTOR},
}

ACTUATED_GC_TENDON_NAMES = []
for tendon_name in GC_TENDONS.keys():
    if not tendon_name.endswith('dp'):
        ACTUATED_GC_TENDON_NAMES.append(tendon_name)

THUMB_TENDON_IDXES = []
FINGER_PIP_IDXES = []
FINGER_MCP_RIGHT_IDXES = []
FINGER_MCP_LEFT_IDXES = []
THUMB_TENDON_NAMES = []
FINGER_PIP_NAMES = []
FINGER_MCP_RIGHT_NAMES = []
FINGER_MCP_LEFT_NAMES = []

for idx, tendon_name in enumerate(ACTUATED_GC_TENDON_NAMES):
    if tendon_name.startswith('thumb'):
        THUMB_TENDON_NAMES.append(tendon_name)
        THUMB_TENDON_IDXES.append(idx + 1)
    else:
        if tendon_name.endswith('mcp'):
            FINGER_MCP_RIGHT_IDXES.append(idx + 1)
            FINGER_MCP_RIGHT_NAMES.append(tendon_name)
        if tendon_name.endswith('pp'):
            FINGER_MCP_LEFT_IDXES.append(idx + 1)
            FINGER_MCP_LEFT_NAMES.append(tendon_name)
        if tendon_name.endswith('mp'):
            FINGER_PIP_IDXES.append(idx + 1)
            FINGER_PIP_NAMES.append(tendon_name)

FINGER_TO_TIP: Dict[str, str] = {
    'thumb': 'thumb_fingertip',
    'index': 'index_fingertip',
    'middle': 'middle_fingertip',
    'ring': 'ring_fingertip',
    'pinky': 'pinky_fingertip',
}

FINGER_TO_BASE = {
    'thumb': 'thumb_cmc_v3_1',
    'index': 'index_mcp_v1_1',
    'middle': 'middle_mcp_v1_1',
    'ring': 'ring_mcp_v1_1',
    'pinky': 'pinky_mcp_v1_1',
}

# TRACKER-SPECIFIC CONSTANTS
# Sensoryx glove
SENSORYX_PALM_OFFSET_ARRAY = [-0.015, 0.015, 0.097]
SENSORYX_ROTATION_OFFSET_ARRAY_RAD = [-1.57, -0.58, -0.15]
SENSORYX_FLIP_AXIS = [False, False, False]

# Manus glove
MANUS_PALM_OFFSET_ARRAY = [-0.015, 0.015, 0.097]
MANUS_ROTATION_OFFSET_ARRAY_RAD = [-1.57, -0.58, -0.25]
MANUS_FLIP_AXIS = [False, False, False]
MANUS_SCALE_Z = 0.9
# TODO (fbenedek) - change this to automatic, using similar method to normal2model_mat
JOINT_MAP_NDARRAY = np.array(
    [
        [1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        [
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            -1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
        ],
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
        ],
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
        ],
    ]
).astype(np.float32)
# the mean-delta representation for MCP angles is simpler to steer when using
# motor-level control, so we're providing utility matrices for it here
MCP_TO_MEAN_DELTA = np.eye(16)
for i in range(4):
    MCP_TO_MEAN_DELTA[3 * i + 4 : 3 * i + 6, 3 * i + 4 : 3 * i + 6] = np.array(
        [[0.5, 0.5], [1.0, -1.0]]
    )

MCP_FROM_MEAN_DELTA = np.eye(16)
for i in range(4):
    MCP_FROM_MEAN_DELTA[3 * i + 4 : 3 * i + 6, 3 * i + 4 : 3 * i + 6] = np.array(
        [[1.0, 0.5], [1.0, -0.5]]
    )


ROBOT_INIT_ANGLE_DEGREES = 0.5
MOTOR_LIMIT_LOWER = -180
MOTOR_LIMIT_HIGHER = 180
MCP_ABD_JOINT_IDXES = [4, 7, 10, 13]
MCP_FLEX_JOINT_IDXES = [5, 8, 11, 14]
MCP_JOINT_IDX_PAIRS = [
    [abd, flex] for (abd, flex) in zip(MCP_ABD_JOINT_IDXES, MCP_FLEX_JOINT_IDXES)
]
MCP_JOINT_IDXES = []
for mcp_pair in MCP_JOINT_IDX_PAIRS:
    MCP_JOINT_IDXES.append(mcp_pair[0])
    MCP_JOINT_IDXES.append(mcp_pair[1])
PIP_JOINT_IDXES = [6, 9, 12, 15]
THUMB_JOINT_IDXES = [0, 1, 2, 3]

# PIP limits by default
# Hardware allows the lower limit to be around -15 deg
# but it looks off so we're clipping it
GC_LIMITS_LOWER = [0] * len(JOINT_NAMES_NORMAL_P04)
GC_LIMITS_UPPER = [95] * len(JOINT_NAMES_NORMAL_P04)
# MCP ABDUCTION limits
for mcp_abd_idx in MCP_ABD_JOINT_IDXES:
    GC_LIMITS_LOWER[mcp_abd_idx] = -15.0
    GC_LIMITS_UPPER[mcp_abd_idx] = 15.0
# MCP limits
for mcp_idx in MCP_FLEX_JOINT_IDXES:
    GC_LIMITS_LOWER[mcp_idx] = -5.0
    GC_LIMITS_UPPER[mcp_idx] = 90.0
# Thumb limits
THUMB_LIMITS_LOWER = [-30, -55, -40, -50]
THUMB_LIMITS_UPPER = [100, 55, 30, 40.0]
for thumb_idx, full_idx in enumerate(THUMB_JOINT_IDXES):
    GC_LIMITS_LOWER[full_idx] = THUMB_LIMITS_LOWER[thumb_idx]
    GC_LIMITS_UPPER[full_idx] = THUMB_LIMITS_UPPER[thumb_idx]


JOINT_NAMES_NORMAL = []
JOINT_NAMES_MODEL = []
jointname2id_dict_model = {}
jointname2id_dict_normal = {}

MOTOR_NAMES = [str(i) for i in range(16)]


def init_matrices():
    global \
        JOINT_NAMES_MODEL, \
        JOINT_NAMES_NORMAL, \
        jointname2id_dict_model, \
        jointname2id_dict_normal, \
        model2normal_mat, \
        normal2model_mat
    JOINT_NAMES_NORMAL = JOINT_NAMES_NORMAL_P04
    JOINT_NAMES_MODEL = JOINT_NAMES_MODEL_P04
    # construct a dict mapping from joint name to joint id, for fast lookup
    for i, joint_name in enumerate(JOINT_NAMES_MODEL):
        jointname2id_dict_model[joint_name] = i
    for i, joint_name in enumerate(JOINT_NAMES_NORMAL):
        jointname2id_dict_normal[joint_name] = i
    # construct matrix that maps between joint conventions
    model2normal_mat = np.zeros((len(JOINT_NAMES_NORMAL), len(JOINT_NAMES_MODEL)))
    # since P04 only uses pin joints, model->normal convention is 1 everywhere
    for i_model, joint_name_model in enumerate(JOINT_NAMES_MODEL):
        if joint_name_model in JOINT_NAMES_NORMAL:
            model2normal_mat[jointname2id_dict_normal[joint_name_model], i_model] = 1.0
    # for the other direction, we'll set the dp joints to 0.5
    normal2model_mat = np.zeros((len(JOINT_NAMES_MODEL), len(JOINT_NAMES_NORMAL)))
    for i_normal, joint_name_normal in enumerate(JOINT_NAMES_NORMAL):
        if joint_name_normal in JOINT_NAMES_MODEL:
            normal2model_mat[jointname2id_dict_model[joint_name_normal], i_normal] = 1.0
        if joint_name_normal.endswith('mp'):
            # if this is the pp2mp joint,
            # the linked mp2dp joint should also be updated
            mp2dp_factor = MP2DP_FACTOR
            dp_joint_name = joint_name_normal.replace('pp2mp', 'mp2dp')
            normal2model_mat[jointname2id_dict_model[dp_joint_name], i_normal] = (
                mp2dp_factor
            )
